@@ -1,3 +1,4 @@
+# -*- coding: utf-8 -*-
 
 """define tracker for tracer use.
 
@@ -13,7 +14,8 @@ from tingyun.armoury.ammunition.timer import Timer
 from tingyun.battlefield.knapsack import knapsack
 from tingyun.logistics.warehouse.database_node import DatabaseNode
 from tingyun.logistics.warehouse.tracker_node import TrackerNode
-from tingyun.logistics.warehouse.error_node import ErrorNode
+from tingyun.logistics.warehouse.error_node import ErrorNode, ExternalErrorNode
+from tingyun.logistics.warehouse.exception_node import ExceptionNode
 from tingyun.packages import six
 
 console = logging.getLogger(__name__)
@@ -30,26 +32,33 @@ class Tracker(object):
         self.framework = framework
         self.proxy = proxy
         self.enabled = False
+        self.rum_enabled = True  # current transaction switch to control auto insert browser agent.
         self._settings = None
 
         self.thread_id = knapsack().current_thread_id()
-        self._tracker_id = id(self)
 
         self.background_task = False
+        self.is_root = True  # 是否为根节点
         self.start_time = 0
         self.end_time = 0
+        self.trace_interval = 0  # 距离上一次
         self.queque_start = 0
         self._queque_time = 0
         self.trace_node = []
         self.async_func_trace_time = 0  # millisecond
 
-        self._errors = []
+        self._errors = ''
+        self._exception = []
+        self.external_error = []
         self._custom_params = {}
         self._slow_sql_nodes = []
 
+        self.http_method = "GET"
         self.request_uri = None
-        self.http_status = 0
+        self.http_status = 500
         self.request_params = {}
+        self.header_params = {}
+        self.cookie_params = {}
         self._priority = 0
         self._group = None
         self._name = None
@@ -62,7 +71,24 @@ class Tracker(object):
         self.thread_name = "Unknown"
         self.referer = ""
 
-        self._trace_guid = None
+        # 作为调用者存储的自己的或者被调者返回的信息
+        self._trace_guid = ""  # 被调用时，若产生慢过程生成该id用于标识此次慢过程
+        self._tingyun_id = ""  # 该实例自身的id
+        self._called_traced_data = {}
+        self.called_external_id = ''
+
+        # 做为被调用者，存储的调用者信息
+        self.call_trace_id = ""  # 调用方产生的trace id
+        self.call_external_id = ""  # 调用方产生的外部调用id
+        self.call_tingyun_id = ""  # 调用方发出的听云id，包含账号id
+        self.call_req_id = ""  # 调用方请求标识
+
+        # 0 indicate it's not called the service.
+        self.db_time = 0
+        self.external_time = 0
+        self.redis_time = 0
+        self.memcache_time = 0
+        self.mongo_time = 0
 
         # set the global switch of agent.  the return application settings is download from server witch is merged
         # with the local settings on server
@@ -78,6 +104,9 @@ class Tracker(object):
                 if self._settings:
                     self.enabled = True
 
+        if self.enabled:
+            self._tingyun_id = self._settings.tingyunIdSecret
+
     def __enter__(self):
         """
         :return:
@@ -93,20 +122,26 @@ class Tracker(object):
             raise
 
         self.start_time = time.time()
+        if self.proxy.previous_time:
+            trace_interval = int(self.start_time * 1000) - self.proxy.previous_time
+            self.trace_interval = trace_interval
+        else:
+            self.trace_interval = 0
+
         self.trace_node.append(Timer(None))
 
-    def __exit__(self, exc_type, exc_val, exc_tb, async=False):
+    def __exit__(self, exc_type, exc_val, exc_tb, ty_async=False):
         """
         :param exc_type:
         :param exc_val:
         :param exc_tb:
-        :param async: in tornado, the tracker is not saved in thread cache
+        :param ty_async: in tornado, the tracker is not saved in thread cache
         :return:
         """
         if not self.enabled:
             return
 
-        if not async:
+        if not ty_async:
             try:
                 self.drop_tracker()
             except Exception as err:
@@ -120,8 +155,10 @@ class Tracker(object):
         if exc_type is not None and exc_val is not None and exc_tb is not None:
             self.record_exception(exc_type, exc_val, exc_tb)
 
+        # 结束时间以应用最后一个字节写入时间为准，在框架结束的地方可能已经处理好了
         self.end_time = time.time()
-        duration = int((self.end_time - self.start_time) * 1000)
+        self.proxy.previous_time = int(self.end_time * 1000)
+        duration = self.duration
 
         root_node = self.trace_node.pop()
         children = root_node.children
@@ -130,7 +167,7 @@ class Tracker(object):
         tracker_type = "WebAction" if not self.background_task else "BackgroundAction"
         group = self._group or ("Python" if self.background_task else "Uri")
         request_params = self.filter_params(self.request_params)
-        uri = self.url_encode(self.request_uri)
+        uri = self.request_uri
 
         # replace the specified metric apdex_t
         apdex_t = self._settings.apdex_t
@@ -138,19 +175,31 @@ class Tracker(object):
         if path in self._settings.action_apdex:
             apdex_t = self._settings.action_apdex.get(path)
 
+        self.finalize_tracker_data()
+
         node = TrackerNode(type=tracker_type, group=group, name=self._name, start_time=self.start_time,
                            end_time=self.end_time, request_uri=uri, duration=duration, thread_name=self.thread_name,
                            http_status=self.http_status, exclusive=exclusive, children=tuple(children),
                            path=self.path, errors=self._errors, apdex_t=apdex_t, queque_time=self.queque_time,
                            custom_params=self._custom_params, request_params=request_params,
-                           referer=self.referer, slow_sql=self._slow_sql_nodes, trace_guid=self.trace_guid)
+                           trace_id=self.call_trace_id,
+                           referer=self.referer, slow_sql=self._slow_sql_nodes, trace_guid=self.generate_trace_guid(),
+                           trace_data=self._called_traced_data, external_error=self.external_error,
+                           trace_interval=self.trace_interval, is_root=self.is_root, exception=self._exception)
 
         self.proxy.record_tracker(node)
 
     def record_exception(self, exc=None, value=None, tb=None, params=None, tracker_type="WebAction",
-                         ignore_errors=[]):
-        """ record the exception for trace
+                         ignore_errors=(), is_error=True, additional_msg=''):
+        """每个url错误只会有一个，无论记录多少次，但异常可以有多个
         :param exc:
+        :param value:
+        :param tb:
+        :param params:
+        :param tracker_type:
+        :param ignore_errors:
+        :param is_error: 用于区分错误和异常
+        :param additional_msg: 用于区分错误和异常
         :return:
         """
         if not self._settings or not self._settings.error_collector.enabled:
@@ -163,51 +212,128 @@ class Tracker(object):
             console.warning("None exception is got. skip it now. %s, %s, %s", exc, value, tb)
             return
 
-        if self.http_status in self._settings.error_collector.ignored_status_codes:
-            console.debug("ignore the status code %s", self.http_status)
-            return
+        m = value.__class__.__module__
+        name = value.__class__.__name__
+        fullname = '%s:%s' % (m, name) if m else name
 
-        # 'True' - ignore the error.
-        # 'False'- record the error.
-        # ignore status code and maximum error number filter will be done in data engine because of voiding repeat count
-        # method ignore_errors() is used to detect the the status code which is can not captured
-        if callable(ignore_errors):
-            should_ignore = ignore_errors(exc, value, tb, self._settings.error_collector.ignored_status_codes)
-            if should_ignore:
+        if is_error:
+            if is_error and self.http_status in self._settings.error_collector.ignored_status_codes:
+                console.debug("ignore the status code %s", self.http_status)
                 return
 
-        # think more about error occurred before deal the status code.
-        if self.http_status in self._settings.error_collector.ignored_status_codes:
-            console.debug("record_exception: ignore  error collector status code")
+            # 'True' - ignore the error.
+            # 'False'- record the error.
+            # ignore status code and maximum error number filter will be done in data engine because of
+            # voiding repeat count
+            # method ignore_errors() is used to detect the the status code which is can not captured
+            if is_error and callable(ignore_errors):
+                if ignore_errors(exc, value, tb, self._settings.error_collector.ignored_status_codes):
+                    return
+
+            # think more about error occurred before deal the status code.
+            if is_error and self.http_status in self._settings.error_collector.ignored_status_codes:
+                console.debug("record_exception: ignore  error collector status code")
+                return
+
+            request_params = self.filter_params(self.request_params)
+            if params:
+                custom_params = dict(request_params)
+                custom_params.update(params)
+            else:
+                custom_params = dict(request_params)
+
+            try:
+                message = str(value)
+            except Exception as _:
+                try:
+                    # Assume JSON encoding can handle unicode.
+                    message = six.text_type(value)
+                except Exception as _:
+                    message = '<unprintable %s object>' % type(value).__name__
+
+            stack_trace = traceback.extract_tb(tb)
+            node = ErrorNode(error_time=int(time.time()), http_status=self.http_status, error_class_name=fullname,
+                             uri=self.request_uri, thread_name=self.thread_name, message=message,
+                             stack_trace=stack_trace, request_params=custom_params, tracker_type=tracker_type,
+                             referer=self.referer)
+            self._errors = node
+        else:
+            message = getattr(value, 'message', '')
+
+            try:
+                message = str(message) if message else str(value)
+            except Exception:
+                message = ''
+
+            if additional_msg:
+                message = '%s (%s)' % (message, additional_msg[: self._settings.max_sql_hash_length])
+
+            stack_trace = traceback.extract_tb(tb) if self._settings and self._settings.exception.stack_enabled else ''
+            node = ExceptionNode(exception_time=int(time.time()), class_name=fullname, tracker_type=tracker_type,
+                                 message=message, stack_trace=stack_trace)
+            self._exception.append(node)
+
+        return node
+
+    def record_external_error(self, url, error_code, http_status=0, _exception=None, request_params=None,
+                              tracker_type="External", module_name="Unknown"):
+        """
+        :return:
+        """
+        ignored_status = [401, ]
+        if http_status in ignored_status or int(http_status) < 400:
+            console.info("Agent caught http status code %s, ignore it now.", http_status)
             return
 
-        module = value.__class__.__module__
-        name = value.__class__.__name__
-        fullname = '%s:%s' % (module, name) if module else name
+        status_code = error_code or http_status
+        error_class_name = ""
+        stack_trace = ""
 
-        request_params = self.filter_params(self.request_params)
-        if params:
-            custom_params = dict(request_params)
-            custom_params.update(params)
-        else:
-            custom_params = dict(request_params)
+        if _exception:
+            exc, value, tb = _exception
+            module = value.__class__.__module__
+            name = value.__class__.__name__
+            error_class_name = '%s:%s' % (module or "Unknown", name or "Unknown")
+            stack_trace = traceback.extract_tb(tb)
 
-        try:
-            message = str(value)
-        except Exception as _:
-            try:
-                # Assume JSON encoding can handle unicode.
-                message = six.text_type(value)
-            except Exception as _:
-                message = '<unprintable %s object>' % type(value).__name__
+        node = ExternalErrorNode(error_time=int(time.time()), status_code=status_code, thread_name=self.thread_name,
+                                 url=url, error_class_name=error_class_name, http_status=http_status,
+                                 stack_trace=stack_trace, request_params=request_params, tracker_type=tracker_type,
+                                 module_name=module_name)
 
-        stack_trace = traceback.extract_tb(tb)
-        node = ErrorNode(error_time=int(time.time()), http_status=self.http_status, error_class_name=fullname,
-                         uri=self.url_encode(self.request_uri), thread_name=self.thread_name, message=message,
-                         stack_trace=stack_trace, request_params=custom_params, tracker_type=tracker_type,
-                         referer=self.referer)
+        self.external_error.append(node)
 
-        self._errors.append(node)
+    def finalize_tracker_data(self):
+        """处理追踪器中需要最好采集的数据，如不能回传的跨应用数据
+        :return:
+        """
+        if not self.call_tingyun_id or not self.settings:
+            return None
+
+        if self.duration < self.settings.action_tracer.action_threshold:
+            return None
+
+        app_id = self.call_tingyun_id[self.call_tingyun_id.find('|') + 1:]
+        code_time = self.duration - self.queque_time - self.db_time - self.external_time - self.redis_time - \
+                    self.memcache_time - self.mongo_time
+        data = {
+            "applicationId": app_id,
+            "transactionId": self.call_trace_id,
+            "externalId": self.call_external_id,
+            "time": {
+                "duration": self.duration,
+                "qu": self.queque_time,
+                "db": self.db_time,
+                "ex": self.external_time,
+                "rds": self.redis_time,
+                "mc": self.memcache_time,
+                "mon": self.mongo_time,
+                "code": 0 if code_time < 0 else code_time
+            }
+        }
+
+        self._custom_params['message.wait'] = self.trace_interval
+        self._custom_params['entryTrace'] = data
 
     @property
     def path(self):
@@ -219,6 +345,18 @@ class Tracker(object):
         tracker_type = "WebAction" if not self.background_task else "BackgroundAction"
         name = self._name or "Undefined"
 
+        try:
+            named_metric = self.settings.naming.naming_web_action(self.http_method, self.request_uri,
+                                                                  self.request_params,
+                                                                  self.header_params, self.cookie_params)
+        except Exception as err:
+            console.error("Error occurred when parsing naming rules. %s", err)
+            named_metric = None
+
+        if named_metric:
+            path = '%s/%s' % (tracker_type, self.url_encode(named_metric))
+            return path
+
         uri_params_captured = self._settings.web_action_uri_params_captured.get(self.request_uri, '')
         if self.request_uri in self._settings.web_action_uri_params_captured:
             match_param = ''
@@ -245,8 +383,8 @@ class Tracker(object):
             return encoded_url
 
         # drop the uri first /
-        uri = self.merge_urls(uri.replace('/', '', 1))
-        encoded_url = uri.replace("/", "%2F")
+        merged_uri = self.merge_urls(uri.replace('/', '', 1))
+        encoded_url = merged_uri.replace("/", "%2F")
 
         return encoded_url
 
@@ -404,17 +542,45 @@ class Tracker(object):
         return self._settings
 
     @property
-    def trace_guid(self):
+    def duration(self):
         """
         :return:
         """
+        end_time = self.end_time
+        if not self.end_time:
+            end_time = time.time()
+
+        return int((end_time - self.start_time) * 1000)
+
+    def generate_trace_guid(self):
+        """
+        :return:
+        """
+        duration = self.duration
+        if duration < self.settings.action_tracer.action_threshold:
+            return ""
+
         if self._trace_guid:
             return self._trace_guid
 
         guid = '%s%s' % (time.time(), random.random())
+        guid = guid.replace('.', '')
         self._trace_guid = guid[:30]
         return self._trace_guid
 
+    def generate_trace_id(self):
+        """ call_trace_id 由调用方生成，若被调方又作为调用方，则复用调用方的call_trace_id
+        :return:
+        """
+
+        if self.call_trace_id:
+            return self.call_trace_id
+
+        trace_id = '%s%s' % (time.time(), random.random())
+        trace_id = trace_id.replace('.', '')
+        self.call_trace_id = trace_id[:16]
+        return self.call_trace_id
+
     @property
     def queque_time(self):
         """
@@ -432,3 +598,29 @@ def current_tracker():
     :return:
     """
     return knapsack().current_tracker()
+
+
+def record_exception(is_error=False):
+    """异常的对象
+    :param is_error:
+    :return:
+    """
+    tracker = current_tracker()
+    if not tracker:
+        return
+
+    ex = tracker.record_exception(is_error=is_error)
+    if not is_error and len(tracker.trace_node) > 0:
+        # relate the exception to the nearest node.
+        node = tracker.trace_node[0]
+        while node:
+            if hasattr(node, "exception") and isinstance(node.exception, list):
+                node.exception.append(ex)
+                break
+
+            if hasattr(node, 'children') and len(node.children) > 0:
+                node = node.children[0]
+            else:
+                break
+
+
