@@ -25,6 +25,19 @@ memcache_attr = {
 
 
 def detect(module):
+    def parse_connect_param(instance, *args, **kwargs):
+        """
+        :return:
+        """
+        host, port = "Unknown", "Unknown"
+        if hasattr(instance, 'servers') and len(instance.servers) > 0:
+            try:
+                host, port = instance.servers[0].ip, instance.servers[0].port
+            except AttributeError:
+                pass
+
+        return host, port
+
     for attr in memcache_attr:
         if hasattr(module.Client, attr):
-            wrap_memcache_trace(module, memcache_attr[attr]['path'], memcache_attr[attr]['command'])
+            wrap_memcache_trace(module, memcache_attr[attr]['path'], attr, server=parse_connect_param)
