@@ -1,7 +1,10 @@
+# -*- coding:utf-8 -*-
+
 """This module implements data recording and reporting for an application
 
 """
 
+import json
 import logging
 import threading
 import time
@@ -19,6 +22,7 @@ from tingyun.packages.wrapt.decorators import synchronized
 from tingyun.logistics.mapper import CONSTANCE_OUT_DATE_CONFIG, CONSTANCE_RETRY_DATA, CONSTANCE_SERVER_UNAVAILABLE
 from tingyun.logistics.mapper import CONSTANCE_DISCARD_DATA, CONSTANCE_HARVEST_ERROR, CONSTANCE_INVALID_DATA_TOKEN
 from tingyun.logistics.mapper import CONSTANCE_LICENSE_INVALID, CONSTANCE_SESSION_NOT_ACTIVED
+from tingyun.logistics.mapper import CONSTANCE_INVALID_LICENSE_KEY, MAX_RETRY_UPLOAD_FAILED_TIMES
 
 
 console = logging.getLogger(__name__)
@@ -35,21 +39,25 @@ class Application(object):
         """
         console.info("Init Application with name %s, linked-name %s", app_name, linked_app)
 
+        # if retry upload data failed. will not rollback the data and reset the session
+        self.retry_upload_times = 0
+
         self._app_name = app_name
         self._linked_app = sorted(set(linked_app))
         self._active_session = None
+        self.connect_thread = None
         self._is_license_valid = True
 
         self.sampler = []
 
-        self.connect_retry_time = 30
+        self.connect_retry_time = 60.0
         self._connect_event = threading.Event()
         self._packets_lock = threading.Lock()  # used to lock the core collect data
         self._packager = Packager()
 
         self._agent_commands_lock = threading.Lock()
         self.profile_manager = get_profile_manger()
-        self.profile_status = False
+        self.profile_status = None
 
         self.__max_tracker = 2000
         self._tracker_count = 0
@@ -90,12 +98,15 @@ class Application(object):
         :return:
         """
         # TODO: do some check for agent status
+        if self.connect_thread and self.connect_thread.is_alive():
+            console.warning("Activate thread is active. maybe something wrong when dispatcher activate application.")
+            return False
+
         self._connect_event.clear()
 
-        connect_thread = threading.Thread(target=self.connect_to_headquarters,
-                                          name="TingYun-session-thread-%s" % int(time.time()))
-        connect_thread.setDaemon(True)
-        connect_thread.start()
+        self.connect_thread = threading.Thread(target=self.connect_to_headquarters, name="tingYunSessionThread")
+        self.connect_thread.setDaemon(True)
+        self.connect_thread.start()
 
         return True
 
@@ -104,33 +115,32 @@ class Application(object):
         settings.
         :return:
         """
-        # TODO: do some check for agent status
-        if self._active_session:
-            console.info("Application is actived, skip to connect to server.")
-            return True
-
         # ensure the main thread get to run first
         time.sleep(0.01)
 
         while not self._active_session:
+            self.retry_upload_times = 0
             settings = global_settings()
+
             try:
                 active_session = create_connection(None, self._app_name, self._linked_app, env_config(), settings)
             except InvalidLicenseException as _:
-                self._is_license_valid = False
                 console.warning("Invalid license in configuration, agent will stop to work please fix license and"
                                 "restart agent again")
-                break
-            except Exception as _:
+
+                self._is_license_valid = False
+                self._connect_event.wait(self.connect_retry_time)
+                continue
+            except Exception:
                 # use the harvest controller signal to control the connection
                 console.exception("Connect to agent server failed, Connection will try again in 1 min.")
                 self._connect_event.wait(self.connect_retry_time)
+                continue
+            finally:
                 if self._connect_event.isSet():
                     console.info("Agent is shutting down, stop the connection to server now.")
                     return
 
-                continue
-
             if active_session:
                 self._is_license_valid = True
                 self._connect_event.set()
@@ -182,7 +192,7 @@ class Application(object):
 
         # controller should ignore session/license error code, because the connecting thread doing/did it
         if not self._active_session:
-            console.info("Application not registered to server yet, skip harvest data.")
+            console.warning("Application not registered to server yet, skip harvest data.")
             return CONSTANCE_SESSION_NOT_ACTIVED, self._app_name, self._linked_app
 
         if not self._is_license_valid:
@@ -194,7 +204,12 @@ class Application(object):
             self._last_tracker = 0.0
             stat = self._packager.packets_snapshot()
 
+        config = self.application_config
+
         for sampler in self.__data_sampler:
+            if not config.enabled:
+                break
+
             try:
                 for s in sampler.metrics(current_harvest - last_harvest):
                     stat.record_time_metrics(s)
@@ -202,37 +217,50 @@ class Application(object):
                 console.exception("Sample the data occurred errors. %s", err)
 
         try:
-            config = self.application_config
-
-            # send metric data
+            # send metric data even if disabled by server
             performance_metric = self.get_performance_metric(stat, last_harvest, current_harvest, config.audit_mode)
             if config.audit_mode:
                 console.info("Agent capture the performance data %s", performance_metric)
 
             result = self._active_session.send_performance_metric(performance_metric)
+
+            # is application is disabled by server. send empty data and ignore all of other data. even if it is not
+            # captured.
+            if not config.enabled:
+                stat.reset_metric_packets()
+                return
+
             self.process_metric_id(result, config.daemon_debug)
 
             if config.error_collector.enabled:
                 error_trace = self.get_error_trace_data(stat)
-                self._active_session.send_error_trace(error_trace)
+                external_error_trace = self.get_external_error_trace(stat)
+                exception_trace = self.get_exception_trace_data(stat)
 
                 if config.audit_mode:
                     console.info("Agent capture the error trace data %s", error_trace)
+                    console.info("Agent capture the external error trace data %s", external_error_trace)
+                    console.info("Agent capture the exception data %s", exception_trace)
+
+                self._active_session.send_error_trace(error_trace)
+                self._active_session.send_external_error_trace(external_error_trace)
+                self._active_session.send_exception_trace(exception_trace)
 
             if config.action_tracer.enabled:
                 slow_action_data = stat.action_trace_data()
-                self._active_session.send_action_trace(slow_action_data)
 
                 if config.audit_mode:
                     console.info("Agent capture the slow action data %s", slow_action_data)
 
+                self._active_session.send_action_trace(slow_action_data)
+
             if config.action_tracer.slow_sql:
                 slow_sql_data = stat.slow_sql_data()
-                self._active_session.send_sql_trace(slow_sql_data)
-
                 if config.audit_mode:
                     console.info("Agent capture the slow sql data %s", slow_sql_data)
 
+                self._active_session.send_sql_trace(slow_sql_data)
+
             stat.reset_metric_packets()
 
             # get the commands and execute it.
@@ -247,25 +275,35 @@ class Application(object):
             ret_code = CONSTANCE_OUT_DATE_CONFIG
             console.info("Config changed in server, reset the connect now.")
         except InvalidDataTokenException as _:
-            # need to reset the connection
+            # if license key when upload the data ,then reset the session
             self._active_session = None
             ret_code = CONSTANCE_INVALID_DATA_TOKEN
             console.info("Data token is valid, register the application %s again now", self._app_name)
-        except DiscardDataForRequest as _:
-            ret_code = CONSTANCE_DISCARD_DATA
-        except ServerIsUnavailable as _:
-            ret_code = CONSTANCE_SERVER_UNAVAILABLE
+        except InvalidLicenseException as _:
+            # if invalid license error occurred when upload data. current data will be lost. And dispatcher will reset
+            # session.
+            self._active_session = None
+            ret_code = CONSTANCE_INVALID_LICENSE_KEY
         except RetryDataForRequest as _:
+            console.warning("This exception indicates server service can not touched. if this error continues. please "
+                            "report to us for further investigation. thank u")
+
+            # if upload try times more than max, reset the session and raise invalid data token, thant will trigger
+            # reconnection to data collector. This will avoid some issue for memory use because of data rollback.
             ret_code = CONSTANCE_RETRY_DATA
+            if self.retry_upload_times >= MAX_RETRY_UPLOAD_FAILED_TIMES:
+                self._active_session = None
+                ret_code = CONSTANCE_INVALID_DATA_TOKEN
 
-            with self._packets_lock:
-                try:
-                    self._packager.rollback(stat)
-                except Exception as err:
-                    console.warning("rollback performance data failed. %s", err)
+                console.warning("Upload data to the dc failed with max retry %s", self.retry_upload_times)
+            else:
+                self.retry_upload_times += 1
+                with self._packets_lock:
+                    try:
+                        self._packager.rollback(stat)
+                    except Exception as err:
+                        console.warning("rollback(%s) performance data failed. %s", self.retry_upload_times, err)
 
-            console.debug("This exception indicates server service can not touched. if this error continues. please "
-                          "report to us for further investigation. thank u")
         except Exception as err:
             ret_code = CONSTANCE_HARVEST_ERROR
             console.exception("Some error occurred in agent code. if this error continues. "
@@ -340,20 +378,34 @@ class Application(object):
         # disable the id mechanism
         metric_name_ids = self.metric_name_ids
         if audit_mode:
-            metric_name_ids = {"actions": {}, "apdex": {}, "components": {}, "general": {}, "errors": {}}
+            metric_name_ids = {"actions": {}, "apdex": {}, "components": {}, "general": {}, "errors": {},
+                               "exceptions": {}}
 
         performance = {
             "type": "perfMetrics",
             "timeFrom": int(last_harvest),
             "timeTo": int(current_harvest),
             "interval": int(current_harvest - last_harvest),
-            "actions": stat.action_metrics(metric_name_ids["actions"]),
-            "apdex": stat.apdex_data(metric_name_ids["apdex"]),
-            "components": stat.component_metrics(metric_name_ids["components"]),
-            "general": stat.general_trace_metric(metric_name_ids["general"]),
-            "errors": stat.error_packets(metric_name_ids["errors"]),
+            "actions": [],
+            "apdex": [],
+            "components": [],
+            "general": [],
+            "errors": stat.error_packets(metric_name_ids.get('errors', {})),
+            "exceptions": stat.exception_packets(metric_name_ids.get('exceptions', {}))
         }
 
+        # if application is disabled. should send some heartbit data.
+        if not self.application_config.enabled:
+            return performance
+
+        performance['actions'] = stat.action_metrics(metric_name_ids.get("actions", {}))
+        performance['apdex'] = stat.apdex_data(metric_name_ids.get("apdex", {}))
+        performance['components'] = stat.component_metrics(metric_name_ids.get("components", {}))
+        performance['general'] = stat.general_trace_metric(metric_name_ids.get("general", {}))
+
+        if 0 != len(self._active_session.config.quantile):
+            performance["config"] = {"nbs.quantile": json.dumps(self._active_session.config.quantile_org)}
+
         return performance
 
     def get_error_trace_data(self, stat):
@@ -366,7 +418,39 @@ class Application(object):
         }
 
         # no error data recorded return None as mark.
-        if 0 == len(stat.error_trace_data()):
+        if 0 == len(error_trace['errors']):
+            error_trace = []
+
+        return error_trace
+
+    def get_exception_trace_data(self, stat):
+        """
+        :param stat:
+        :return:
+        """
+        exception_trace = {
+            "type": "exceptionTraceData",
+            "exceptions": stat.exception_trace_data()
+        }
+
+        if 0 == len(exception_trace['exceptions']):
+            exception_trace = []
+
+        return exception_trace
+
+    def get_external_error_trace(self, stat):
+        """
+        :param stat:
+        :return:
+        """
+        error_data = stat.external_error_data()
+        error_trace = {
+            "type": "externalErrorTraceData",
+            "errors": error_data
+        }
+
+        # no error data recorded return None as mark.
+        if 0 == len(error_data):
             error_trace = []
 
         return error_trace
@@ -440,7 +524,6 @@ class Application(object):
             return
 
         self.profile_manager.shutdown(stop_cmd=True)
-        self._active_session.send_profile_data({"id": cmd_id, "result": {}})
         self.profile_status = None
         console.info("Stopping profile for application %s", self._app_name)
 
