@@ -1,10 +1,13 @@
-"""this module is implement to deal the tracker all data
+# -*- coding:utf-8 -*-
 
+"""this module is implement to deal the tracker all data
 """
+
 import copy
 
 import logging
 import threading
+from tingyun.logistics.workshop.quantile import QuantileP2
 from tingyun.logistics.workshop.packets import TimePackets, ApdexPackets, SlowSqlPackets
 from tingyun.packages import six
 
@@ -16,21 +19,29 @@ class Packager(object):
     """
     """
     def __init__(self):
-        self.__max_error_count = 10
         self.__settings = None
 
         self.__time_packets = {}  # store time packets data key is name + scope
         self.__apdex_packets = {}
         self.__action_packets = {}
         self.__general_packets = {}
-        # it's maybe include {name: {trackerType, count, $ERROR_COUNTERS_ITEM}}
+        # it's maybe include {name: {trackerType, count, $ERROR_ITEM}}
         # $ERROR_ITEM  detail, check the documentation
         self.__traced_errors = {}
+        self.__traced_exception = {}
+
+        # it's maybe include {name: {count, $ERROR_ITEM}}
+        # $ERROR_ITEM  detail, check the documentation
+        self.__traced_external_errors = {}
 
         # format: {$"metric_name": [$slow_action_data, $duration]}
         self.__slow_action = {}
         self.__slow_sql_packets = {}
 
+        # quantile
+        self.__quantile_data = None
+        self.__quantile = {}
+
         self._packets_lock = threading.Lock()
 
     @property
@@ -61,6 +72,7 @@ class Packager(object):
         self.__action_packets = {}
         self.__general_packets = {}
         self.__traced_errors = {}
+        self.__traced_external_errors = {}
         self.__slow_action = {}
         self.__slow_sql_packets = {}
 
@@ -77,73 +89,105 @@ class Packager(object):
         threshold = self.__settings.action_tracer.action_threshold
 
         self.record_time_metrics(tracker.time_metrics())  # deal for component, include framework/db/other
+        self.record_quantile(*tracker.quantile())
         self.record_action_metrics(tracker.action_metrics())  # deal for user action
         self.record_apdex_metrics(tracker.apdex_metrics())  # for recording the apdex
-        self.record_traced_errors(tracker.traced_error())  # for error trace detail
-        self.record_slow_action(tracker.slow_action_trace(node_limit, threshold))
-        self.record_slow_sql(tracker.slow_sql_nodes())
+
+        # 必须放在exception和errors前面执行，因为slowAction最后一段需要统计错误和异常的总数
+        if self.__settings.action_tracer.enabled:
+            self.record_slow_action(tracker.slow_action_trace(node_limit, threshold))
+
+        if self.__settings.error_collector.enabled:
+            self.record_traced_errors(tracker.traced_error())  # for error trace detail
+            self.record_traced_exception(tracker.traced_exception())
+            self.record_traced_external_trace(tracker.traced_external_error())
+
+        if self.__settings.action_tracer.slow_sql:
+            self.record_slow_sql(tracker.slow_sql_nodes())
+
+    def record_quantile(self, action, duration):
+        """
+        :return:
+        """
+        if 0 == len(self.__settings.quantile):
+            return
+
+        self.__quantile_data = (action, duration)
 
     def record_slow_sql(self, nodes):
         """
         :param nodes: the slow sql node
         :return:
         """
-        if not self.__settings.action_tracer.slow_sql:
-            return
-
         for node in nodes:
-            key = node.identifier
+            # key = node.request_uri + node.sql
+            key = node.metric
             packets = self.__slow_sql_packets.get(key)
-            if not packets and len(self.__slow_sql_packets) < self.__settings.slow_sql_count:
+            if not packets:
                 packets = SlowSqlPackets()
-                self.__slow_sql_packets[key] = packets
 
-            if packets:
-                packets.merge_slow_sql_node(node)
+            packets.merge_slow_sql_node(node)
+            self.__slow_sql_packets[key] = packets
 
     def record_slow_action(self, slow_action):
         """
         :param slow_action:
         :return:
         """
-        if not self.__settings.action_tracer.enabled:
-            return
-
-        threshold = self.__settings.action_tracer.action_threshold
-        top_n = self.__settings.action_tracer.top_n
-        if slow_action[1] < threshold:
-            return
-
-        if len(self.__slow_action) > top_n:
-            console.debug("The slow action trace is reach the top, action(%s) is ignored.", slow_action[2])
+        # 不能在这里通过判断阈值是否到达临界去除法慢过程，因为可能因为跨应用而强制触发
+        # 如果节点返回`无效数据`则直接丢弃
+        if not slow_action:
             return
 
         metric_name = slow_action[2]
         if metric_name not in self.__slow_action:
-            self.__slow_action[metric_name] = [slow_action, slow_action[1]]
+            self.__slow_action[metric_name] = [[slow_action, slow_action[1]]]
         else:
-            if slow_action[1] > self.__slow_action[metric_name][1]:
-                self.__slow_action[metric_name] = [slow_action, slow_action[1]]
+            # every request url/metric can save top_n action trace.
+            self.__slow_action[metric_name].append([slow_action, slow_action[1]])
 
     def record_traced_errors(self, traced_errors):
         """
         :return:
         """
-        if not self.__settings.error_collector.enabled:
-            return
-
         for error in traced_errors:
-            if len(self.__traced_errors) > self.__max_error_count:
-                console.debug("Error trace is reached maximum limitation.")
-                break
-
             if error.error_filter_key in self.__traced_errors:
                 self.__traced_errors[error.error_filter_key]["count"] += 1
-                self.__traced_errors[error.error_filter_key]["item"][-3] += 1
+                self.__traced_errors[error.error_filter_key]["item"][-4] += 1
             else:
-                self.__traced_errors[error.error_filter_key] = {"count": 1, "item": error.trace_data,
+                self.__traced_errors[error.error_filter_key] = {"count": 1,
+                                                                "item": error.trace_data,
                                                                 "tracker_type": error.tracker_type}
 
+    def record_traced_exception(self, traced_exception):
+        """
+        :param traced_exception:
+        :return:
+        """
+        for ex in traced_exception:
+            if ex.filter_key in self.__traced_exception:
+                self.__traced_exception[ex.filter_key]['count'] += 1
+                self.__traced_exception[ex.filter_key]['item'][-4] += 1
+            else:
+                self.__traced_exception[ex.filter_key] = {
+                    "count": 1, "item": ex.trace_data, "tracker_type": ex.tracker_type
+                }
+
+    def record_traced_external_trace(self, traced_errors):
+        """
+        :param traced_errors:
+        :return:
+        """
+        for error in traced_errors:
+            if error.error_filter_key in self.__traced_external_errors:
+                self.__traced_external_errors[error.error_filter_key]["count"] += 1
+                self.__traced_external_errors[error.error_filter_key]["item"][-3] += 1
+
+            else:
+                self.__traced_external_errors[error.error_filter_key] = {"count": 1, "status_code": error.status_code,
+                                                                         "item": error.trace_data,
+                                                                         "tracker_type": error.tracker_type}
+
     def record_general_metric(self, metric):
         """
         :param metric:
@@ -173,7 +217,7 @@ class Packager(object):
         key = (metric.name, metric.scope or '')  # metric key for protocol
         packets = self.__time_packets.get(key)
         if packets is None:
-            packets = TimePackets()
+            packets = getattr(metric, 'packets', None) or TimePackets()
 
         packets.merge_time_metric(metric)
         self.__time_packets[key] = packets
@@ -281,8 +325,23 @@ class Packager(object):
             else:
                 packets["count"] += value["count"]
 
+        for key, value in six.iteritems(stat.__traced_exception):
+            packets = self.__traced_exception.get(key)
+            if not packets:
+                self.__traced_exception[key] = copy.copy(value)
+            else:
+                packets["count"] += value["count"]
+
+        for key, value in six.iteritems(stat.__traced_external_errors):
+            packets = self.__traced_external_errors.get(key)
+            if not packets:
+                self.__traced_external_errors[key] = copy.copy(value)
+            else:
+                packets["count"] += value["count"]
+
     def merge_metric_packets(self, snapshot):
-        """
+        """从慢过程以及其他错误信用发生的概率来看，慢过程、错误属于小概率事件，所以不在数据接入时过滤，
+        而是在合并时判断是否超过限制
         :param snapshot:
         :return:
         """
@@ -309,10 +368,40 @@ class Packager(object):
 
         # TODO: think more about the background task
         for key, value in six.iteritems(snapshot.__traced_errors):
+            if len(self.__traced_errors.get(key, [])) > self.__settings.max_error_trace:
+                console.debug("Error trace is reached maximum limitation. %s", self.__settings.max_error_trace)
+                continue
+
             packets = self.__traced_errors.get(key)
             if not packets:
                 self.__traced_errors[key] = copy.copy(value)
             else:
+                packets["item"][-4] += value["count"]
+                packets["count"] += value["count"]
+
+        for key, value in six.iteritems(snapshot.__traced_exception):
+            if len(self.__traced_exception.get(key, [])) > self.__settings.exception.max_type_count:
+                console.debug("Exception trace is reached maximum limitation. %s",
+                              self.__settings.exception.max_type_count)
+                continue
+
+            packets = self.__traced_exception.get(key)
+            if not packets:
+                self.__traced_exception[key] = copy.copy(value)
+            else:
+                packets["item"][-4] += value["count"]
+                packets["count"] += value["count"]
+
+        for key, value in six.iteritems(snapshot.__traced_external_errors):
+            if len(self.__traced_external_errors.get(key, [])) > self.__settings.max_error_trace:
+                console.debug("External error trace is reached maximum limitation. %s", self.__settings.max_error_trace)
+                continue
+
+            packets = self.__traced_external_errors.get(key)
+            if not packets:
+                self.__traced_external_errors[key] = copy.copy(value)
+            else:
+                packets["item"][-3] += value["count"]
                 packets["count"] += value["count"]
 
         # generate general data
@@ -323,32 +412,40 @@ class Packager(object):
             else:
                 packets.merge_packets(value)
 
-        # for slow action
-        top_n = self.__settings.action_tracer.top_n
+        # for action trace
+        top_n = self.__settings.action_tracer.max_action_trace_per_action
         for key, value in six.iteritems(snapshot.__slow_action):
-            if len(self.__slow_action) > top_n:
-                console.debug("The slow action trace count is reach the top.")
+
+            # although the target action trace value is `list`, but it only has 1 element in one metric.
+            if len(self.__slow_action.get(key, [])) > top_n:
+                console.debug("The action trace is reach the top(%s), action(%s) is ignored.", top_n, key)
                 break
 
-            slow_acton = self.__slow_action.get(key)
-            if not slow_acton:
-                self.__slow_action[key] = value[0]
+            if key not in self.__slow_action:
+                self.__slow_action[key] = value
             else:
-                if slow_acton[1] < value[1]:
-                    self.__slow_action[key] = value[0]
+                self.__slow_action[key].extend(value)
 
         # for slow sql
         max_sql = self.__settings.slow_sql_count
         for key, value in six.iteritems(snapshot.__slow_sql_packets):
             if len(self.__slow_sql_packets) > max_sql:
-                console.debug("The slow sql trace count is reach the top.")
+                console.debug("Slow sql count is more than max count %s ",  max_sql)
+                continue
 
             slow_sql = self.__slow_sql_packets.get(key)
             if not slow_sql:
                 self.__slow_sql_packets[key] = value
             else:
-                if value.slow_sql_node.duration > slow_sql.slow_sql_node.duration:
-                    self.__slow_sql_packets[key] = value
+                slow_sql.merge_packets(value)
+
+        # for quantile
+        if 0 != len(self.__settings.quantile):
+            action, duration = snapshot.__quantile_data
+            if action not in self.__quantile:
+                self.__quantile[action] = QuantileP2(self.__settings.quantile)
+
+            self.__quantile[action].add(duration)
 
     def reset_metric_packets(self):
         """
@@ -360,9 +457,13 @@ class Packager(object):
         self.__general_packets = {}
 
         self.__traced_errors = {}
+        self.__traced_external_errors = {}
         self.__slow_action = {}
         self.__slow_sql_packets = {}
 
+        self.__quantile = {}
+        self.__traced_exception = {}
+
     def packets_snapshot(self):
         """
         :return:
@@ -373,9 +474,12 @@ class Packager(object):
         self.__action_packets = {}
         self.__apdex_packets = {}
         self.__traced_errors = {}
+        self.__traced_external_errors = {}
         self.__general_packets = {}
         self.__slow_action = {}
         self.__slow_sql_packets = {}
+        self.__quantile = {}
+        self.__traced_exception = {}
 
         return stat
 
@@ -393,6 +497,7 @@ class Packager(object):
             upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
             result.append([upload_key, value])
 
+        self.__time_packets = {}
         return result
 
     def apdex_data(self, metric_name_ids):
@@ -407,6 +512,8 @@ class Packager(object):
             upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
             result.append([upload_key, value])
 
+        # reset the data if returned for upload.
+        self.__apdex_packets = {}
         return result
 
     def action_metrics(self, metric_name_ids):
@@ -418,33 +525,59 @@ class Packager(object):
             upload_key = {"name": key[0]}
             upload_key_str = '%s' % key[0]
             upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
-            result.append([upload_key, value])
 
+            if 0 != len(self.__settings.quantile):
+                result.append([upload_key, value, self.__quantile[upload_key_str].markers])
+            else:
+                result.append([upload_key, value])
+
+        # reset the data if returned for upload.
+        self.__action_packets = {}
         return result
 
     def error_packets(self, metric_name_ids):
         """stat the error trace metric for performance
         :return:
         """
+        external_error, web_action_error = 'External', 'WebAction'
+        error_types = [external_error, web_action_error]
         error_count = {
             "Errors/Count/All": 0,
             "Errors/Count/AllWeb": 0,
             "Errors/Count/AllBackground": 0
         }
 
-        for error_filter_key, error in six.iteritems(self.__traced_errors):
-            error_count["Errors/Count/All"] += error["count"]
+        def parse_error_trace(traced_data):
+            for error_filter_key, error in six.iteritems(traced_data):
+                error_count["Errors/Count/All"] += error["count"]
 
-            if error["tracker_type"] == "WebAction":
-                error_count["Errors/Count/AllWeb"] += error["count"]
+                if error["tracker_type"] in error_types:
+                    error_count["Errors/Count/AllWeb"] += error["count"]
+                    filter_keys = error_filter_key.split("_|")
 
-                action_key = "Errors/Count/%s" % error_filter_key.split("_|")[0]
-                if action_key not in error_count:
-                    error_count[action_key] = error["count"]
+                    action_key = "Errors/Count/%s" % filter_keys[0]
+                    if action_key not in error_count:
+                        error_count[action_key] = error["count"]
+                    else:
+                        error_count[action_key] += error["count"]
+
+                    error_type_key = "Errors/Type:%s/%s" % (filter_keys[2], filter_keys[0])
+                    if error_type_key not in error_count:
+                        error_count[error_type_key] = error["count"]
+                    else:
+                        error_count[error_type_key] += error["count"]
+
+                    if error["tracker_type"] == external_error:
+                        action_key = "Errors/Type:%s/%s" % (error["status_code"], filter_keys[0])
+                        if action_key not in error_count:
+                            error_count[action_key] = error["count"]
+                        else:
+                            error_count[action_key] += error["count"]
                 else:
-                    error_count[action_key] += error["count"]
-            else:
-                error_count["Errors/Count/AllBackground"] += 1
+                    error_count["Errors/Count/AllBackground"] += 1
+
+        parse_error_trace(self.__traced_errors)
+        parse_error_trace(self.__traced_external_errors)
 
         stat_value = []
         for key, value in six.iteritems(error_count):
@@ -455,13 +588,72 @@ class Packager(object):
 
         return stat_value
 
+    def exception_packets(self, metric_name_ids):
+        """
+        :param metric_name_ids:
+        :return:
+        """
+        exception_count = {
+            "Exception/Count/All": 0,
+            "Exception/Count/AllWeb": 0,
+            "Exception/Count/AllBackground": 0
+        }
+
+        for ex_filter_key, ex in six.iteritems(self.__traced_exception):
+            exception_count["Exception/Count/All"] += ex["count"]
+            filter_keys = ex_filter_key.split("_|")
+
+            if ex["tracker_type"] == 'WebAction':
+                exception_count["Exception/Count/AllWeb"] += ex["count"]
+
+                action_key = "Exception/Count/%s" % filter_keys[0]
+                if action_key not in exception_count:
+                    exception_count[action_key] = ex["count"]
+                else:
+                    exception_count[action_key] += ex["count"]
+
+                exception_type_key = "Exception/Type:%s/%s" % (filter_keys[1], filter_keys[0])
+                if exception_type_key not in exception_count:
+                    exception_count[exception_type_key] = ex["count"]
+                else:
+                    exception_count[exception_type_key] += ex["count"]
+            else:
+                exception_count["Exception/Count/AllBackground"] += 1
+
+        stat_value = []
+        for key, value in six.iteritems(exception_count):
+            upload_key = {"name": key}
+            upload_key_str = '%s' % key
+            upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
+            stat_value.append([upload_key, [value]])
+
+        return stat_value
+
     # stat for error trace data
     # rely to the basic error trace data structure
     def error_trace_data(self):
         """
         :return:
         """
-        return [error["item"] for error in six.itervalues(self.__traced_errors)]
+        rtv = [error["item"] for error in six.itervalues(self.__traced_errors)]
+        self.__traced_errors = {}
+
+        return rtv
+
+    def exception_trace_data(self):
+        """
+        :return:
+        """
+        rtv = [ex["item"] for ex in six.itervalues(self.__traced_exception)]
+        self.__traced_errors = {}
+
+        return rtv
+
+    def external_error_data(self):
+        rtv = [error["item"] for error in six.itervalues(self.__traced_external_errors)]
+        self.__traced_external_errors = {}
+
+        return rtv
 
     def general_trace_metric(self, metric_name_ids):
         """
@@ -474,6 +666,7 @@ class Packager(object):
             upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
             result.append([upload_key, value])
 
+        self.__general_packets = {}
         return result
 
     def action_trace_data(self):
@@ -483,7 +676,13 @@ class Packager(object):
         if not self.__slow_action:
             return []
 
-        return {"type": "actionTraceData", "actionTraces": [value for value in six.itervalues(self.__slow_action)]}
+        trace_data = []
+
+        for traces in six.itervalues(self.__slow_action):
+            for trace in traces:
+                trace_data.append(trace[0])
+
+        return {"type": "actionTraceData", "actionTraces": trace_data}
 
     def slow_sql_data(self):
         """
@@ -511,4 +710,5 @@ class Packager(object):
                                         node.total_call_time, node.max_call_time,
                                         node.min_call_time, str(params)])
 
+        self.__slow_sql_packets = {}
         return result
